@@ -25,6 +25,15 @@ DEFAULT_REPOSITORIES = {
 ui = termui.UI()
 
 
+def choices(*choices: str) -> Callable[[str], str]:
+    def converter(value: str) -> str:
+        if value not in choices:
+            raise ValueError(f"Value must be one of {choices}, got {value!r}")
+        return value
+
+    return converter
+
+
 def load_config(file_path: Path) -> dict[str, Any]:
     """Load a nested TOML document into key-value pairs
 
@@ -131,6 +140,12 @@ class Config(MutableMapping[str, str]):
             "PDM_USE_UV",
             ensure_boolean,
         ),
+        "lock.format": ConfigItem(
+            "The format of the lock file, can be `pdm` or `pylock`",
+            "pdm",
+            env_var="PDM_LOCK_FORMAT",
+            coerce=choices("pdm", "pylock"),
+        ),
         "global_project.fallback": ConfigItem(
             "Use the global project implicitly if no local project is found",
             False,
@@ -294,6 +309,12 @@ class Config(MutableMapping[str, str]):
             self.get_defaults() if is_global else {},
         )
 
+    def reload(self) -> None:
+        """Reload the config from file"""
+        new_data = load_config(self.config_file)
+        self._file_data.clear()
+        self._file_data.update(new_data)
+
     def load_theme(self) -> rich.theme.Theme:
         if not self.is_global:  # pragma: no cover
             raise PdmUsageError("Theme can only be loaded from global config")
