@@ -63,7 +63,7 @@ from pdm.models.requirements import (
 from pdm.models.session import PDMPyPIClient
 from pdm.project.config import Config
 from pdm.project.core import Project
-from pdm.utils import find_python_in_path, normalize_name, parse_version, path_to_url
+from pdm.utils import normalize_name, parse_version
 
 if TYPE_CHECKING:
     from typing import Protocol
@@ -304,8 +304,9 @@ def build_env_wheels() -> Iterable[Path]:
 
 
 @pytest.fixture(autouse=True)
-def temp_env() -> Generator[MutableMapping[str, str]]:
+def temp_env(monkeypatch: pytest.MonkeyPatch) -> Generator[MutableMapping[str, str]]:
     old_env = os.environ.copy()
+    monkeypatch.setattr("pdm.models.candidates.PreparedCandidate._build_dir_cache", {})
     try:
         yield os.environ
     finally:
@@ -328,7 +329,7 @@ def build_env(build_env_wheels: Iterable[Path], tmp_path_factory: pytest.TempPat
     p = Core().create_project(d)
     env = PythonEnvironment(p, prefix=str(d), python=sys.executable)
     for wheel in build_env_wheels:
-        install_wheel(wheel, env)
+        install_wheel(wheel, env, requested=True)
     return d
 
 
@@ -406,9 +407,7 @@ def project_no_init(
     tmp_path.joinpath("caches").mkdir(parents=True)
     p.global_config["cache_dir"] = tmp_path.joinpath("caches").as_posix()
     p.global_config["log_dir"] = tmp_path.joinpath("logs").as_posix()
-    python_path = find_python_in_path(sys.base_prefix)
-    if python_path is None:
-        raise ValueError("Unable to find a Python path")
+    python_path = Path(sys.executable)
     p._saved_python = python_path.as_posix()
     monkeypatch.delenv("VIRTUAL_ENV", raising=False)
     monkeypatch.delenv("CONDA_PREFIX", raising=False)
@@ -501,12 +500,11 @@ def local_finder_artifacts() -> Path:
 
 @pytest.fixture
 def local_finder(project_no_init: Project, local_finder_artifacts: Path) -> None:
-    artifacts_dir = str(local_finder_artifacts)
     project_no_init.pyproject.settings["source"] = [
         {
             "type": "find_links",
             "verify_ssl": False,
-            "url": path_to_url(artifacts_dir),
+            "url": local_finder_artifacts.as_uri(),
             "name": "pypi",
         }
     ]
