@@ -34,11 +34,9 @@ if TYPE_CHECKING:
     from re import Match
     from typing import IO, Any, Iterator
 
-    from pdm._types import FileHash, RepositoryConfig
+    from pdm._types import FileHash, HiddenText, RepositoryConfig
     from pdm.compat import Distribution
 
-_egg_fragment_re = re.compile(r"(.*)[#&]egg=[^&]*")
-
 try:
     _packaging_version = importlib_metadata.version("packaging")
 except Exception:
@@ -189,7 +187,8 @@ def url_to_path(url: str) -> str:
 
     WINDOWS = sys.platform == "win32"
 
-    assert url.startswith("file:"), f"You can only turn file: urls into filenames (not {url!r})"
+    if not url.startswith("file:"):
+        raise ValueError(f"You can only turn file: urls into filenames (not {url!r})")
 
     _, netloc, path, _, _ = parse.urlsplit(url)
 
@@ -220,16 +219,10 @@ def url_to_path(url: str) -> str:
     return path
 
 
-def path_to_url(path: str) -> str:
-    """
-    Convert a path to a file: URL.  The path will be made absolute and have
-    quoted path parts.
-    """
-    from urllib.request import pathname2url
-
-    path = os.path.normpath(os.path.abspath(path))
-    url = parse.urljoin("file:", pathname2url(path))
-    return url
+def split_path_fragments(path: Path) -> tuple[Path, str]:
+    """Split a path into fragments"""
+    left, sep, right = path.as_posix().partition("#egg=")
+    return Path(left), sep + right
 
 
 def expand_env_vars(credential: str, quote: bool = False, env: Mapping[str, str] | None = None) -> str:
@@ -244,7 +237,7 @@ def expand_env_vars(credential: str, quote: bool = False, env: Mapping[str, str]
         env = os.environ
 
     def replace_func(match: Match) -> str:
-        rv = env.get(match.group(1), match.group(0))
+        rv = env.get(match.group(1), "")
         return parse.quote(rv, "") if quote else rv
 
     return re.sub(r"\$\{(.+?)\}", replace_func, credential)
@@ -448,14 +441,6 @@ def is_pip_compatible_with_python(python_version: Version | str) -> bool:
     return requires_python.contains(python_version, True)
 
 
-def path_without_fragments(path: str) -> Path:
-    """Remove egg fragment from path"""
-    match = _egg_fragment_re.search(path)
-    if not match:
-        return Path(path)
-    return Path(match.group(1))
-
-
 def is_in_zipapp() -> bool:
     """Check if the current process is running in a zipapp"""
     return not os.path.exists(__file__)
@@ -582,3 +567,16 @@ def get_requirement_from_override(name: str, value: str) -> str:
         else:
             req = f"{name}{value}"
     return req
+
+
+def hide_url(url: str) -> HiddenText:
+    """Redact the URL for display purposes."""
+    from pdm._types import HiddenText
+
+    parsed = parse.urlsplit(url)
+    if "@" not in parsed.netloc:
+        return HiddenText(url, url)
+    *_, netloc = parsed.netloc.rpartition("@")
+    netloc = f"*****@{netloc}"
+    redacted = parse.urlunsplit((parsed.scheme, netloc, parsed.path, parsed.query, parsed.fragment))
+    return HiddenText(url, redacted)
